/*
    $Id: pathutil.c,v 1.4.2.1 2002/01/03 15:36:18 soulshock Exp $

    Portions Copyright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

static const char _rcsid[] = "$Id: pathutil.c,v 1.4.2.1 2002/01/03 15:36:18 soulshock Exp $";

#include "pathutil.h"
#include "cdpath.h"
#include "site_vfs.h"

#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>

#include <glib.h>

/* regexp based pathname expression simplifier */

#define RE_COUNT 4
static regex_t _re[RE_COUNT];
static const gchar *_re_str[RE_COUNT][2] =
  {
    {"//+", "/"},
    {"/\\./", "/"},
    {"/[^/]+/\\.\\./", "/"},
    {"^/\\.\\./", "/"}
  };

static void
_init_regex (void)
{
  static gboolean _init = FALSE;

  if (!_init)
    {
      gint i;

      for (i = 0; i < RE_COUNT; i++)
	if (regcomp (&_re[i], _re_str[i][0], REG_EXTENDED))
	  g_assert_not_reached ();

      _init = TRUE;
    }
}

static gboolean
_substitute (GString *str, gint idx)
{
  regmatch_t rm;
  gint pos, len;

  if (regexec (&_re[idx], str->str, 1, &rm, 0))
    return FALSE;

  pos = rm.rm_so;
  len = rm.rm_eo - rm.rm_so;

  g_string_erase (str, pos, len);
  g_string_insert (str, pos, _re_str[idx][1]);

  return TRUE;
}

/* TODO 
 * make use of GCache for memo(r)izing results,
 * since this function is supposed to be called often
 */

gchar *
path_simplify (const gchar pathname[])
{
  GString *str;
  gchar *result;
  gint idx;

  _init_regex ();

  str = g_string_new (pathname);

  g_string_prepend_c (str, '/');
  g_string_append_c (str, '/');

  for (idx = 0; idx < RE_COUNT; idx++)
    while (_substitute (str, idx));

  if (str->len > 1 && str->str[str->len - 1] == '/')
    g_string_truncate (str, str->len - 1);

  result = str->str;
  g_string_free (str, FALSE);

  return result;
}

gchar *
path_add (const gchar basepath[], const gchar addpath[])
{
  gchar *result = NULL;

  g_return_val_if_fail (basepath != NULL, NULL);
  g_return_val_if_fail (addpath != NULL, NULL);

  if (addpath[0] == '/')	/* absolute path */
    result = path_simplify (addpath);
  else
    {
      gchar *tmp = g_strconcat ("/", basepath, "/", addpath, NULL);
      result = path_simplify (tmp);
      g_free (tmp);
    }

  g_assert (result != NULL);

  return result;
}

gboolean 
path_child_p (const gchar parent[], const gchar child[])
{
  gchar *_parent;
  gchar *_child;
  gboolean result = FALSE;

  _parent = path_simplify (parent);
  _child = path_simplify (child);

  if (!strncmp (_parent, _child, strlen (_parent)))
    result = TRUE;

  g_free (_parent);
  g_free (_child);

  return result;
}
  
/* old stuff */

/* new emulation wrapper using path_add */

int add_dir (const gchar *src, const gchar *dir, gchar *dest, gint maxlen)
{
  gchar *newpath = NULL;

  if (!dir)
    dir = "/";
  
  newpath = path_add (src, dir);

  /* g_printerr ("add_dir %s %s %s\n", src, dir, newpath); */

  if (strlen (newpath) >= maxlen)
    {
      errno = ENAMETOOLONG;
      g_free (newpath);
      return -1;
    }

  strncpy (dest, newpath, maxlen);

  g_free (newpath);

  return 0;
}

gboolean path_in_cdpath (const gchar pathname[])
{
  gchar *_path;

  if ((_path = path_add (vfs_getcwd (), pathname)))
    {
      const gchar *tmp = cdpath_match (_path);

      g_free (_path);

      return tmp ? TRUE : FALSE;
   }

  return FALSE;
}
